// Title:	BiDiPro DLL module
// Desc:	SPI flash
// Ver:		0.0.6
// Found:	26-09-2008
// Last:	01-06-2009
// CPU:		386+
// Model:	Tiny
// Compil:	Borland C++ 3.1 (DOS)
// Author:	Oleg O. Chukaev aka 0leg

#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <limits.h>
#include "..\dlldata.h"
#include "..\errors.h"
#include "..\bidi.h"


#define dlp ((DLLDATA *) NULL)		/* ⥫    */
#define bdport dlp->bdprt		/* 㪠⥫  hardware ࠩ */

struct CHDEFS {
	dword size;	// 
	byte adrmode;	//  樨
};

CHDEFS chd[]={
//	size        adrmode
	{0x000001L, 0},		//1
	{0x008000L, 0},		//256
	{0x010000L, 0},		//512
	{0x020000L, 0},		//010
	{0x040000L, 0},		//020
	{0x080000L, 0},		//040
	{0x100000L, 0},		//080
	{0x200000L, 0},		//016
};

// ७   CFG:
static struct CFGKEY cfgkeys[] = {
	{'D', "id_alg", 0, ULONG_MAX, 20},		//Read ID mode
	{'D', "wrsr_alg", 0, ULONG_MAX, 20},		//WRSR alg.
	{'D', "er_alg", 0, ULONG_MAX, 20},		//Erase mode
	{'D', "aai", 1, ULONG_MAX, 20},			//AAI mode (0..2)
	{'D', "wpage", 0, ULONG_MAX, 65536},		//Page size
	{'D', "wtime", 20, ULONG_MAX, 100000L},		//Write time (us)
	{'D', "etime", 100000L, ULONG_MAX, 40000000L},	//Erase time (us)
	{0, NULL, 0, 0},				//End of table
};

#define cfg_id_alg	cfgkeys[0].data
#define cfg_wrsr_alg	cfgkeys[1].data
#define cfg_er_alg	cfgkeys[2].data
#define cfg_aai		cfgkeys[3].data
#define cfg_wpage	cfgkeys[4].data
#define cfg_wtime	cfgkeys[5].data
#define cfg_etime	cfgkeys[6].data

#define PIN_CE		0x01
#define PIN_WP		0x04
#define PIN_SI		0x08
#define PIN_SCK		0x10
#define PIN_HOLD	0x20
#define PIN_SO		0x01

#define SPI_READ	0x03
#define SPI_SERASE	0x20
#define SPI_BERASE	0x52
#define SPI_CERASE	0x60
#define SPI_CERASE1	0x62
#define SPI_CERASE2	0xc7
#define SPI_PRG		0x02
#define SPI_AAI1	0xaf
#define SPI_AAI2	0xad
#define SPI_RDSR	0x05
#define SPI_EWSR	0x50
#define SPI_WRSR	0x01
#define SPI_WREN	0x06
#define SPI_WRDI	0x04
#define SPI_RDID	0xab
#define SPI_RDID1	0x15
#define SPI_RDID2	0x90
#define SPI_RDID_JEDEC	0x9f


static char *copyright="25F* SPI flash module (c) Oleg O. Chukaev  ["__DATE__"]";
static word chtype = 0xFFFF;	//  ROM
static CHDEFS *pchd;
static byte pin_wp = 0;


void power_off (void) {
	bdport (PRST);
}


int power_on (void) {

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	bdport (PFORCE | PWR1 | PAIN);		// DD1:IOO
	bdport (PFORCE | PWR2);			// DD4:OOO
	dlp->voltage (dlp->Vcc, 0, 0);
	bdport (PWB2 | 0x80);			// Vcc pin 36
	bdport (PWA2 | BITTER);
	bdport (PWB1 | PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (10000);			//delay
	return (0);
}


// get value of 1 pin: 0 / 1
static byte get_pin (byte pin) {
	return (bdport (PRA1) & pin) ? 1 : 0;
}


// get values of all pins
static byte get_pins (void) {
	return (bdport (PRA1));
}


// set values of all pins
static void set_pins (byte pins) {
	bdport (PWB1 | pins);
}


// ਥ 
static byte read_spi_byte (void) {

	word i;
	byte d = 0;

	for (i = 0; i < 8; i++) {
		d <<= 1;
		d |= get_pin (PIN_SO);
                set_pins (pin_wp | PIN_HOLD | PIN_SCK);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);		/* SCK=0 */
	}
	return (d);
}


// । 
static int write_spi_byte (byte d) {

	word i;
	byte z;

	for (i = 0; i < 8; i++) {
		z = (d & 0x80) ? PIN_SI : 0;
		set_pins (pin_wp | PIN_HOLD | z);		/* data */
		set_pins (pin_wp | PIN_HOLD | PIN_SCK | z);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);			/* SCK=0 */
		d <<= 1;
	}
	return (0);
}


// unprotect blocks
void unprotect (void) {

	set_pins (PIN_HOLD | pin_wp);
	if (cfg_wrsr_alg == 0) {
		write_spi_byte (SPI_EWSR);
	} else { //if (cfg_wrsr_alg == 1) {
		write_spi_byte (SPI_WREN);
	}
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (1000);
	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WRSR);
	write_spi_byte (0x00);		/* clear BP0, BP1, ... */
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (150000L);	/* for Atmel, Spansion and PMC */
}


// read
int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	byte dat;

        // Status Register
        if (pchd->size == 1) {
		power_on ();
		set_pins (PIN_HOLD | pin_wp);
		write_spi_byte (SPI_RDSR);
		dat = read_spi_byte ();
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
		power_off ();
		dlp->setbytebuf (0, dat);
		return (0);
        }

	power_on ();
	set_pins (PIN_HOLD);
	write_spi_byte (SPI_READ);
	write_spi_byte (0x00);		/* A23..A16 */
	write_spi_byte (0x00);		/* A15..A8 */
	write_spi_byte (0x00);		/* A7..A0 */
	for (adr = 0; adr < pchd->size; adr++) {
		dat = read_spi_byte ();
		dlp->setbytebuf (adr, dat);
		if (adr % istep == 0 && dlp->iprogress (adr, 2))
			break;
	}
	power_off ();
	return (0);
}


// compare
int compare_all (void) {

	dword i;
	dword errors = 0;
	dword istep = pchd->size / 64L;
	byte dat;

	power_on ();
	set_pins (PIN_HOLD);
	write_spi_byte (SPI_READ);
	write_spi_byte (0x00);		/* A23..A16 */
	write_spi_byte (0x00);		/* A15..A8 */
	write_spi_byte (0x00);		/* A7..A0 */
	for (i = 0; i < pchd->size; i++) {
		dat = read_spi_byte ();
		if (dlp->getbytebuf (i) != dat) {
			/* ⬥ ᮢ */
			dlp->setbmark (i, 1);
			errors++;
		}
		if (i % istep == 0 && dlp->iprogress (i, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return (0);
}


// blank check
int blank_chk (void) {

	dword i;
	dword istep = pchd->size / 64L;

	power_on ();
	set_pins (PIN_HOLD);
	write_spi_byte (SPI_READ);
	write_spi_byte (0x00);		/* A23..A16 */
	write_spi_byte (0x00);		/* A15..A8 */
	write_spi_byte (0x00);		/* A7..A0 */
	for (i = 0; i < pchd->size; i++) {
		if (read_spi_byte () != 0xff) {
			dlp->adr = i;
			power_off ();
			return (1);
		}
		if (i % istep == 0) {
			if (dlp->iprogress (i, 2))
				break;
		}
	}
	power_off ();
	return (0);
}


// read id
/*
0 (SST):	0xab, 0x00, 0x00, 0x00, read mfr, read dev
1 (Atmel 1):	0x15, read mfr, read dev
2 (Atmel 2):	0xab, read mfr, read dev_h, read dev_l
3 (JEDEC):	0x9f, read mfr, read type, read capacity
4 (SST, EON):	0x90, 0x00, 0x00, 0x00, read mfr, read dev
*/
int read_id (void) {

	dword id1, id2;

	dlp->edata = 0;
	power_on ();
	set_pins (PIN_HOLD);
	switch (cfg_id_alg) {
	case 0:
		write_spi_byte (SPI_RDID);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);		/* ID addr (0: mfr_id first) */
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		break;

	case 1:
		write_spi_byte (SPI_RDID1);
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		break;

	case 2:
		write_spi_byte (SPI_RDID);
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		id2 <<= 8;
		id2 |= read_spi_byte ();
		break;

	case 3:
		write_spi_byte (SPI_RDID_JEDEC);
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		id2 <<= 8;
		id2 |= read_spi_byte ();
		break;

	case 4:
		write_spi_byte (SPI_RDID2);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);		/* ID addr (0: mfr_id first) */
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		break;
	}
	power_off ();
	dlp->edata = id1 | (id2 << 16L);
	return (0);
}


// erase
/*
0 (SST, Atmel 2):	0x60
1 (Atmel 1):		0x62
2 (SST 2, Atmel, EON):	0xC7
*/
int erase (void) {

	dword istep, i, w0;

	pin_wp = PIN_WP;
	power_on ();
	unprotect ();
	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WREN);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	set_pins (PIN_HOLD | pin_wp);
	if (cfg_er_alg == 0) {
		write_spi_byte (SPI_CERASE);
	} else if (cfg_er_alg == 1) {
		write_spi_byte (SPI_CERASE1);
	} else if (cfg_er_alg == 2) {
		write_spi_byte (SPI_CERASE2);
	}
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	if (cfg_etime == 0L) {				/* poll BUSY bit */
		dlp->mcsdelay (10000);
		dlp->iprogress (1600, 0);		/* init, 16 s */
		istep = 1600 / 64;
		set_pins (PIN_HOLD | pin_wp);
		write_spi_byte (SPI_RDSR);
		i = 0;
		do {
			dlp->mcsdelay (10000);
			w0 = read_spi_byte ();
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			i++;
		} while ((w0 & 0x01) == 0x01);
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
	} else if (cfg_etime < 300000L) {		/* 0.3 s */
		dlp->mcsdelay (cfg_etime);		/* delay */
	} else {
		dlp->iprogress (cfg_etime/1000, 0);	/* init */
		istep = cfg_etime / 64000L;
		for (i = 0; i < cfg_etime/1000; i++) {
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			dlp->mcsdelay (1000);		/* delay */
		}
		dlp->iprogress (0, 1);			/* close */
	}
		dlp->iprogress (0, 1);			/* close */
	power_off ();
	pin_wp = 0;
	return (0);
}


// write
/*
AAI/Page mode:
0 (Atmel, PMC):	0x02, addr2, addr1, addr0, data0, ..., data128(256)
1 (SST):	0xaf, addr2, addr1, addr0, data0, WAIT, 0xaf, data1, ...
2 (SST):	0xad, addr2, addr1, addr0, data0, data1, WAIT,
		0xad, data2, data3, ...
*/
int write_all (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size/64L;

        // Status Register
        if (pchd->size == 1) {
		pin_wp = PIN_WP;
		power_on ();
		set_pins (PIN_HOLD | pin_wp);
		write_spi_byte (SPI_WREN);
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
		dlp->mcsdelay (1000);
		set_pins (PIN_HOLD | pin_wp);
		write_spi_byte (SPI_WRSR);
		write_spi_byte (dlp->getbytebuf (0));
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
		dlp->mcsdelay (150000L);/* for Atmel, Spansion and PMC */
		power_off ();
		pin_wp = 0;
		return (0);
	}

	pin_wp = PIN_WP;
	power_on ();
	unprotect ();
	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WREN);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	set_pins (PIN_HOLD | pin_wp);
	switch (cfg_aai) {
	case 0:
		for (adr = 0; adr < pchd->size; adr += cfg_wpage) {
			write_spi_byte (SPI_PRG);
			write_spi_byte ((adr >> 16L) & 0xff);	/* A23..A16 */
			write_spi_byte ((adr >> 8L) & 0xff);	/* A15..A8 */
			write_spi_byte (adr & 0xff);		/* A7..A0 */
			for (i = 0; i < cfg_wpage; i++) {
				write_spi_byte (dlp->getbytebuf (adr+i));
			}
			set_pins (PIN_HOLD | pin_wp | PIN_CE);
			dlp->mcsdelay (cfg_wtime * cfg_wpage);
			set_pins (PIN_HOLD | pin_wp);
			write_spi_byte (SPI_WREN);
			set_pins (PIN_HOLD | pin_wp | PIN_CE);
			set_pins (PIN_HOLD | pin_wp);
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 1:
		write_spi_byte (SPI_AAI1);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
		write_spi_byte (dlp->getbytebuf (0));
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
		dlp->mcsdelay (cfg_wtime);
		set_pins (PIN_HOLD | pin_wp);
		for (adr = 1; adr < pchd->size; adr++) {
			write_spi_byte (SPI_AAI1);
			write_spi_byte (dlp->getbytebuf (adr));
			set_pins (PIN_HOLD | pin_wp | PIN_CE);
			dlp->mcsdelay (cfg_wtime);
			set_pins (PIN_HOLD | pin_wp);
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 2:
		write_spi_byte (SPI_AAI2);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
		write_spi_byte (dlp->getbytebuf (0));
		write_spi_byte (dlp->getbytebuf (1));
		set_pins (PIN_HOLD | pin_wp | PIN_CE);
		dlp->mcsdelay (cfg_wtime);
		set_pins (PIN_HOLD | pin_wp);
		for (adr = 2; adr < pchd->size; adr += 2) {
			write_spi_byte (SPI_AAI2);
			write_spi_byte (dlp->getbytebuf (adr));
			write_spi_byte (dlp->getbytebuf (adr+1));
			set_pins (PIN_HOLD | pin_wp | PIN_CE);
			dlp->mcsdelay (cfg_wtime);
			set_pins (PIN_HOLD | pin_wp);
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;
	}

	write_spi_byte (SPI_WRDI);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	power_off ();
	pin_wp = 0;
	return (0);
}


// test
int test (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w1 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	unprotect ();

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w2 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WREN);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w3 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return (0);
}


// lock
int lock_chip (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w1 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	set_pins (PIN_HOLD | pin_wp);
	if (cfg_wrsr_alg == 0) {
		write_spi_byte (SPI_EWSR);
	} else { //if (cfg_wrsr_alg == 1) {
		write_spi_byte (SPI_WREN);
	}
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (1000);
	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WRSR);
	write_spi_byte (0x0c);		/* clear BP0, BP1, ... */
	set_pins (PIN_HOLD | pin_wp | PIN_CE);
	dlp->mcsdelay (150000L);	/* for Atmel, Spansion and PMC */

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w2 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_WREN);
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	set_pins (PIN_HOLD | pin_wp);
	write_spi_byte (SPI_RDSR);
	w3 = read_spi_byte ();
	set_pins (PIN_HOLD | pin_wp | PIN_CE);

	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return (0);
}


// ⠭ ⨯ ROM
word settype (void) {

	word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize) {
			if (chd[i].size == 1) {
				dlp->funmask[DFN_FCOMPARE] = 0;
				dlp->funmask[DFN_READID] = 0;
				dlp->funmask[DFN_ERASE] = 0;
				dlp->funmask[DFN_FBLANK] = 0;
				dlp->funmask[DFN_TEST] = 0;
				dlp->funmask[DFN_LOCK] = 0;
			}
			return (chtype = i);
                }
	}
	return (chtype = 0xFFFF);
}


//----------------------------------------------------------

// main code
extern "C" int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		// ⨢஢ 㯭 㭪樨:
		dlp->funmask[DFN_INIT] = 1;
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->funmask[DFN_FCOMPARE] = 1;
		dlp->funmask[DFN_READID] = 1;
		dlp->funmask[DFN_ERASE] = 1;
		dlp->funmask[DFN_FBLANK] = 1;
		dlp->funmask[DFN_TEST] = 1;
		dlp->funmask[DFN_LOCK] = 1;
		dlp->defstate = 0xFF;
		dlp->copyright = copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		return (0);

	case DFN_INIT:
		if (settype () == 0xFFFF)
			return (ERC_CHIPUNSP);
		pchd = &chd[chtype];
		dlp->dippins = 8;
		return (0);

	case DFN_FREAD:
		return (read_all ());

	case DFN_FWRITE:
		return (write_all ());

	case DFN_FCOMPARE:
		return (compare_all ());

	case DFN_FBLANK:
		return (blank_chk ());

	case DFN_READID:
		return (read_id ());

	case DFN_ERASE:
		return (erase ());

	case DFN_TEST:
		return (test ());

	case DFN_LOCK:
		return (lock_chip ());

	default:
		return (ERC_FUNUNSP);
	}
}

/* E0F */

